"""
Entity associations.
"""

from __future__ import annotations

import weakref
from abc import ABC, abstractmethod
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, Self, cast, final, overload, override
from urllib.parse import quote

from betty.importlib import fully_qualified_name, import_any
from betty.json.linked_data import LinkedDataDumper
from betty.json.schema import Array, Null, OneOf, Schema
from betty.locale.localizable import resolve_localizable
from betty.model import Entity, persistent_id
from betty.model.collections import (
    EntityCollection,
    MultipleTypesEntityCollection,
    SingleTypeEntityCollection,
)
from betty.model.schema import ToManySchema, ToZeroOrOneSchema

if TYPE_CHECKING:
    from ty_extensions import Intersection

    from betty.locale.localizable import ResolvableLocalizable
    from betty.portable import PortableData
    from betty.project import Project


async def _generate_associate_url(project: Project, associate: Entity, /) -> str | None:
    if not persistent_id(associate):
        return None
    if not associate.plugin().public_facing:
        return None
    url_generator = await project.url_generator
    return url_generator.generate(
        f"betty-static:///{associate.plugin().id}/{quote(associate.id)}/index.json"
    )


class AssociationRequired[OwnerT: Entity](RuntimeError):
    """
    Raised when an operation cannot be performed because the association in question is required.
    """

    def __init__(self, association: _Association[OwnerT, Any], owner: OwnerT, /):
        super().__init__(
            f"Association {fully_qualified_name(association.owner_type)}.{association.owner_attr_name} is required, but missing for {owner}."
        )


class _Resolver[T](ABC):
    @abstractmethod
    def resolve(self) -> T:
        """
        Return the resolved entity or entities.

        :raises ResolutionError: Raised if resolution failed.
        """


class ToZeroOrOneResolver[EntityT: Entity](_Resolver[EntityT | None]):
    """
    An object that can optionally resolve to an entity.
    """


class ToOneResolver[EntityT: Entity](_Resolver[EntityT]):
    """
    An object that can resolve to an entity.
    """


class ToManyResolver[EntityT: Entity](_Resolver[Iterable[EntityT]]):
    """
    An object that can resolve to a collection of entities.
    """


class _TemporaryResolver[T](_Resolver[T]):
    @override
    def resolve(self) -> T:
        raise RuntimeError(
            "This temporary resolver was supposed to be replaced. It intentionally cannot resolve itself."
        )


class TemporaryToZeroOrOneResolver[EntityT: Entity](
    _TemporaryResolver[EntityT], ToZeroOrOneResolver[EntityT]
):
    """
    A 'temporary' to-zero-or-one resolver.

    This is helpful to satisfy association requirements in multiple steps. Users **MUST** ensure that this resolver
    is replaced by a real value, because the resolver will never be able to resolve itself.
    """


class TemporaryToOneResolver[EntityT: Entity](
    _TemporaryResolver[EntityT], ToOneResolver[EntityT]
):
    """
    A 'temporary' to-one resolver.

    This is helpful to satisfy association requirements in multiple steps. Users **MUST** ensure that this resolver
    is replaced by a real value, because the resolver will never be able to resolve itself.
    """


class TemporaryToManyResolver[EntityT: Entity](
    _TemporaryResolver[EntityT], ToManyResolver[EntityT]
):
    """
    A 'temporary' to-many resolver.

    This is helpful to satisfy association requirements in multiple steps. Users **MUST** ensure that this resolver
    is replaced by a real value, because the resolver will never be able to resolve itself.
    """


class _Association[OwnerT: Entity, AssociateT: Entity](LinkedDataDumper[OwnerT]):
    _owner_type: type[OwnerT]
    _owner_attr_name: str
    _internal_owner_attr_name: str

    def __init__(
        self,
        associate_type_name: str,
        *,
        label: ResolvableLocalizable,
        description: ResolvableLocalizable | None = None,
        linked_data_embedded: bool = False,
    ):
        self._associate_type_name = associate_type_name
        self._linked_data_embedded = linked_data_embedded
        self._label = resolve_localizable(label)
        self._description = (
            None if description is None else resolve_localizable(description)
        )

    def __set_name__(self, owner: type[OwnerT], name: str) -> None:
        self._owner_type = owner
        self._owner_attr_name = name
        self._internal_owner_attr_name = f"_{name}"
        AssociationRegistry._register(self)

    @override
    def __hash__(self) -> int:
        return hash(
            (
                type(self),
                self._owner_type,
                self._owner_attr_name,
                self._associate_type_name,
                self._linked_data_embedded,
                self._label,
                self._description,
            )
        )

    @property
    def owner_type(self) -> type[OwnerT]:
        """
        The type of the owning entity that contains this association.

        This may be an abstract class.
        """
        return self._owner_type

    @property
    def owner_attr_name(self) -> str:
        """
        The name of the attribute on the owning entity that contains this association.
        """
        return self._owner_attr_name

    @property
    def associate_type(self) -> type[AssociateT]:
        """
        The type of any associate entities.

        This may be an abstract class.
        """
        return cast(
            "type[AssociateT]",
            import_any(self._associate_type_name),
        )

    @abstractmethod
    def resolve(self, owner: OwnerT, /) -> None:
        """
        Resolve any associates the owner may have for this association.
        """

    @abstractmethod
    def associate(self, owner: OwnerT, associate: AssociateT, /) -> None:
        """
        Associate two entities.
        """

    @abstractmethod
    def disassociate(self, owner: OwnerT, associate: AssociateT, /) -> None:
        """
        Disassociate two entities.

        :raises AssociationRequired: Raised if the association is required and the disassociation would leave it without
            any associates.
        """

    @abstractmethod
    def get_associates(self, owner: OwnerT, /) -> Iterable[AssociateT]:
        """
        Get the associates for the given owner.
        """


class _ToOneAssociation[OwnerT: Entity, AssociateT: Entity](
    _Association[OwnerT, AssociateT]
):
    @override
    def associate(self, owner: OwnerT, associate: AssociateT, /) -> None:
        self.__set__(owner, associate)

    @override
    def disassociate(self, owner: OwnerT, associate: AssociateT, /) -> None:
        setattr(owner, self._internal_owner_attr_name, None)

    @overload
    def __get__(self, instance: None, owner: type[OwnerT]) -> Self:
        pass

    @overload
    def __get__(self, instance: OwnerT, owner: type[OwnerT]) -> AssociateT:
        pass

    def __get__(self, instance: OwnerT | None, owner: type[OwnerT]):
        if instance is None:
            return self
        try:
            value = getattr(instance, self._internal_owner_attr_name)
        except AttributeError:
            raise AssociationRequired(self, instance) from None
        else:
            if value is None:
                raise AssociationRequired(self, instance)
            assert not isinstance(value, _Resolver)
            return cast(AssociateT, value)

    def __set__(self, instance: OwnerT, value: ToOneAssociate[AssociateT]) -> None:
        setattr(instance, self._internal_owner_attr_name, value)

    @override
    def get_associates(self, owner: OwnerT, /) -> Iterable[AssociateT]:
        yield self.__get__(owner, type(owner))

    @override
    async def linked_data_schema_for(self, project: Project, /) -> Schema:
        if self._linked_data_embedded:
            return await self.associate_type.linked_data_schema(project)
        # We must allow for the associate to be missing, for example if it has a generated entity ID and the linked data
        # is not embedded, no URL can be generated.
        return ToZeroOrOneSchema(
            title=self._label,
            description=None if self._description is None else self._description,
        )

    @override
    async def dump_linked_data_for(
        self, project: Project, target: Intersection[OwnerT, Entity], /
    ) -> PortableData:
        associate = self.__get__(target, type(target))
        if self._linked_data_embedded:
            return await associate.dump_linked_data(project)
        return await _generate_associate_url(project, associate)


class _ToZeroOrOneAssociation[OwnerT: Entity, AssociateT: Entity](
    _Association[OwnerT, AssociateT]
):
    @override
    def associate(self, owner: OwnerT, associate: AssociateT, /) -> None:
        self.__set__(owner, associate)

    @override
    def disassociate(self, owner: OwnerT, associate: AssociateT, /) -> None:
        if associate == self.__get__(owner, type(owner)):
            self.__delete__(owner)

    @overload
    def __get__(self, instance: None, owner: type[OwnerT]) -> Self:
        pass

    @overload
    def __get__(self, instance: OwnerT, owner: type[OwnerT]) -> AssociateT | None:
        pass

    def __get__(self, instance: OwnerT | None, owner: type[OwnerT]):
        if instance is None:
            return self
        try:
            value = getattr(instance, self._internal_owner_attr_name)
        except AttributeError:
            setattr(instance, self._internal_owner_attr_name, None)
            return None
        else:
            assert not isinstance(value, _Resolver)
            return cast(AssociateT | None, value)

    def __set__(
        self, instance: OwnerT, value: ToZeroOrOneAssociate[AssociateT]
    ) -> None:
        setattr(instance, self._internal_owner_attr_name, value)

    def __delete__(self, instance: OwnerT) -> None:
        self.__set__(instance, None)

    @override
    def get_associates(self, owner: OwnerT, /) -> Iterable[AssociateT]:
        associate = self.__get__(owner, type(owner))
        if associate is not None:
            yield associate

    @override
    async def linked_data_schema_for(self, project: Project, /) -> Schema:
        if self._linked_data_embedded:
            return OneOf(
                await self.associate_type.linked_data_schema(project),
                Null(),
                title=self._label,
                description=None if self._description is None else self._description,
            )
        return ToZeroOrOneSchema(
            title=self._label,
            description=None if self._description is None else self._description,
        )

    @override
    async def dump_linked_data_for(
        self, project: Project, target: Intersection[OwnerT, Entity], /
    ) -> PortableData:
        associate = self.__get__(target, type(target))
        if associate is None:
            return None
        if self._linked_data_embedded:
            return await associate.dump_linked_data(project)
        return await _generate_associate_url(project, associate)


class _ToManyAssociation[
    OwnerT: Entity,
    AssociateT: Entity,
    EntityCollectionT: EntityCollection[Any],
](
    _Association[OwnerT, AssociateT],
):
    @abstractmethod
    def _new_collection(self, instance: OwnerT, /) -> EntityCollectionT:
        pass

    @overload
    def __get__(self, instance: None, owner: type[OwnerT]) -> Self:
        pass

    @overload
    def __get__(self, instance: OwnerT, owner: type[OwnerT]) -> EntityCollectionT:
        pass

    def __get__(self, instance: OwnerT | None, owner: type[OwnerT]):
        if instance is None:
            return self
        try:
            value = getattr(instance, self._internal_owner_attr_name)
        except AttributeError:
            value = self._new_collection(instance)
            setattr(instance, self._internal_owner_attr_name, value)
            return value
        else:
            assert not isinstance(value, _Resolver)
            return cast(EntityCollectionT, value)

    def __set__(self, instance: OwnerT, value: ToManyAssociates[AssociateT]) -> None:
        if isinstance(value, _Resolver):
            setattr(instance, self._internal_owner_attr_name, value)
        else:
            self.__get__(instance, type(instance)).replace(*value)

    def __delete__(self, instance: OwnerT) -> None:
        self.__get__(instance, type(instance)).clear()

    @override
    def associate(self, owner: OwnerT, associate: AssociateT, /) -> None:
        self.__get__(owner, type(owner)).add(associate)

    @override
    def disassociate(self, owner: OwnerT, associate: AssociateT, /) -> None:
        self.__get__(owner, type(owner)).remove(associate)

    @override
    def get_associates(self, owner: OwnerT, /) -> Iterable[AssociateT]:
        yield from self.__get__(owner, type(owner))

    @override
    def resolve(self, owner: OwnerT, /) -> None:
        value = getattr(owner, self._internal_owner_attr_name, None)
        if isinstance(value, _Resolver):
            collection = self._new_collection(owner)
            setattr(owner, self._internal_owner_attr_name, collection)
            collection.add(*value.resolve())

    @override
    async def linked_data_schema_for(self, project: Project, /) -> Schema:
        if self._linked_data_embedded:
            return Array(
                await self.associate_type.linked_data_schema(project),
                title=self._label,
                description=None if self._description is None else self._description,
            )
        return ToManySchema(
            title=self._label,
            description=None if self._description is None else self._description,
        )

    @override
    async def dump_linked_data_for(
        self, project: Project, target: Intersection[OwnerT, Entity], /
    ) -> PortableData:
        associates = self.__get__(target, type(target))
        if self._linked_data_embedded:
            return [
                await associate.dump_linked_data(project) for associate in associates
            ]
        return list(
            filter(
                None,
                [
                    await _generate_associate_url(project, associate)
                    for associate in associates
                ],
            )
        )


class _BidirectionalAssociation[OwnerT: Entity, AssociateT: Entity](
    _Association[OwnerT, AssociateT]
):
    def __init__(
        self,
        associate_type_name: str,
        associate_attr_name: str,
        *,
        label: ResolvableLocalizable,
        linked_data_embedded: bool = False,
        description: ResolvableLocalizable | None = None,
    ):
        self._associate_attr_name = associate_attr_name
        super().__init__(
            associate_type_name,
            label=label,
            description=description,
            linked_data_embedded=linked_data_embedded,
        )

    @override
    def __hash__(self) -> int:
        return hash((super().__hash__(), self._associate_attr_name))

    @property
    def associate_attr_name(self) -> str:
        """
        The association's attribute name on the associate type.
        """
        return self._associate_attr_name

    def inverse(self) -> _BidirectionalAssociation[AssociateT, OwnerT]:
        """
        Get the inverse association.
        """
        association = AssociationRegistry.get_association(
            self.associate_type, self.associate_attr_name
        )
        assert isinstance(association, _BidirectionalAssociation)
        return association


@final
class BidirectionalToZeroOrOne[OwnerT: Entity, AssociateT: Entity](
    _ToZeroOrOneAssociation[OwnerT, AssociateT],
    _BidirectionalAssociation[OwnerT, AssociateT],
):
    r"""
    A bidirectional \\*-to-zero-or-one entity type association.
    """

    @override
    def __set__(
        self, instance: OwnerT, value: ToZeroOrOneAssociate[AssociateT]
    ) -> None:
        previous_associate = self.__get__(instance, type(instance))
        if previous_associate == value:
            return
        super().__set__(instance, value)
        if previous_associate is not None:
            self.inverse().disassociate(previous_associate, instance)
        if not isinstance(value, _Resolver) and value is not None:
            self.inverse().associate(value, instance)

    @override
    def resolve(self, owner: OwnerT, /) -> None:
        value = getattr(owner, self._internal_owner_attr_name, None)
        if isinstance(value, _Resolver):
            associate = value.resolve()
            setattr(owner, self._internal_owner_attr_name, value.resolve())
            if associate:
                self.inverse().associate(associate, owner)


@final
class BidirectionalToOne[OwnerT: Entity, AssociateT: Entity](
    _ToOneAssociation[OwnerT, AssociateT],
    _BidirectionalAssociation[OwnerT, AssociateT],
):
    r"""
    A bidirectional \\*-to-one entity type association.
    """

    @override
    def resolve(self, owner: OwnerT, /) -> None:
        value = getattr(owner, self._internal_owner_attr_name, None)
        if value is None:
            raise AssociationRequired(self, owner)
        if isinstance(value, _Resolver):
            associate = value.resolve()
            setattr(owner, self._internal_owner_attr_name, associate)
            self.inverse().associate(associate, owner)

    @override
    def __set__(self, instance: OwnerT, value: ToOneAssociate[AssociateT]) -> None:
        try:
            previous_associate = cast(
                "AssociateT | None", getattr(self, self._internal_owner_attr_name)
            )
        except AttributeError:
            previous_associate = None
        if previous_associate == value:
            return
        super().__set__(instance, value)
        if previous_associate:
            self.inverse().disassociate(previous_associate, instance)
        if not isinstance(value, _Resolver):
            self.inverse().associate(value, instance)


@final
class BidirectionalToManySingleType[OwnerT: Entity, AssociateT: Entity](
    _ToManyAssociation[OwnerT, AssociateT, SingleTypeEntityCollection[AssociateT]],
    _BidirectionalAssociation[OwnerT, AssociateT],
):
    r"""
    A bidirectional \\*-to-many entity type association where all associates are of the same entity type.
    """

    @override
    def _new_collection(
        self, instance: OwnerT, /
    ) -> SingleTypeEntityCollection[AssociateT]:
        return _BidirectionalSingleTypeAssociateCollection(instance, self)


@final
class BidirectionalToManyMultipleTypes[OwnerT: Entity, AssociateT: Entity](
    _ToManyAssociation[OwnerT, AssociateT, MultipleTypesEntityCollection[AssociateT]],
    _BidirectionalAssociation[OwnerT, AssociateT],
):
    r"""
    A bidirectional \\*-to-many entity type association where associates may be of different entity types.
    """

    @override
    def _new_collection(
        self, instance: OwnerT, /
    ) -> MultipleTypesEntityCollection[AssociateT]:
        return _BidirectionalMultipleTypesAssociateCollection(
            instance,
            self,
        )


@final
class UnidirectionalToZeroOrOne[OwnerT: Entity, AssociateT: Entity](
    _ToZeroOrOneAssociation[OwnerT, AssociateT]
):
    """
    A unidirectional to-zero-or-one entity type association.
    """

    @override
    def resolve(self, owner: OwnerT, /) -> None:
        value = getattr(owner, self._internal_owner_attr_name, None)
        if isinstance(value, _Resolver):
            setattr(owner, self._internal_owner_attr_name, value.resolve())


@final
class UnidirectionalToOne[OwnerT: Entity, AssociateT: Entity](
    _ToOneAssociation[OwnerT, AssociateT]
):
    """
    A unidirectional to-one entity type association.
    """

    @override
    def resolve(self, owner: OwnerT, /) -> None:
        value = getattr(owner, self._internal_owner_attr_name, None)
        if value is None:
            raise AssociationRequired(self, owner)
        if isinstance(value, _Resolver):
            setattr(owner, self._internal_owner_attr_name, value.resolve())


@final
class UnidirectionalToManySingleType[OwnerT: Entity, AssociateT: Entity](
    _ToManyAssociation[OwnerT, AssociateT, SingleTypeEntityCollection[AssociateT]]
):
    """
    A unidirectional to-many entity type association where all associates are of the same entity type.
    """

    @override
    def _new_collection(
        self, instance: OwnerT, /
    ) -> SingleTypeEntityCollection[AssociateT]:
        return SingleTypeEntityCollection[AssociateT]()


@final
class UnidirectionalToManyMultipleTypes[OwnerT: Entity, AssociateT: Entity](
    _ToManyAssociation[OwnerT, AssociateT, MultipleTypesEntityCollection[AssociateT]],
):
    """
    A unidirectional to-many entity type association where associates may be of different entity types.
    """

    @override
    def _new_collection(
        self, instance: OwnerT, /
    ) -> MultipleTypesEntityCollection[AssociateT]:
        return MultipleTypesEntityCollection[AssociateT]()


@final
class AssociationRegistry:
    """
    Inspect any known entity type associations.
    """

    _associations = set[_Association[Any, Any]]()

    @classmethod
    def get_all_associations(
        cls, owner: type | object, /
    ) -> set[_Association[Any, Any]]:
        """
        Get all associations for an owner.
        """
        owner_type = owner if isinstance(owner, type) else type(owner)
        return {
            association
            for association in cls._associations
            if association.owner_type in owner_type.__mro__
        }

    @classmethod
    def get_association[OwnerT: Entity](
        cls, owner: type[OwnerT] | OwnerT, owner_attr_name: str, /
    ) -> _Association[OwnerT, Any]:
        """
        Get the association for a given owner and attribute name.
        """
        for association in cls.get_all_associations(owner):
            if association.owner_attr_name == owner_attr_name:
                return association
        raise ValueError(
            f"No association exists for {fully_qualified_name(owner if isinstance(owner, type) else type(owner))}.{owner_attr_name}."
        )

    @classmethod
    def _register(cls, association: _Association[Any, Any], /) -> None:
        cls._associations.add(association)


class _BidirectionalAssociateCollection[AssociateT: Entity, OwnerT: Entity](
    EntityCollection[AssociateT]
):
    def __init__(
        self,
        owner: OwnerT,
        association: _BidirectionalAssociation[OwnerT, AssociateT],
        /,
    ):
        super().__init__()
        self._association = association
        self.__owner = weakref.ref(owner)

    @property
    def _owner(self) -> OwnerT:
        owner = self.__owner()
        assert owner is not None, (
            "This associate collection's owner no longer exists in memory."
        )
        return owner

    @override
    def _on_add(self, *entities: AssociateT) -> None:
        super()._on_add(*entities)
        for associate in entities:
            self._association.inverse().associate(associate, self._owner)

    @override
    def _on_remove(self, *entities: AssociateT) -> None:
        super()._on_remove(*entities)
        for associate in entities:
            self._association.inverse().disassociate(associate, self._owner)


class _BidirectionalSingleTypeAssociateCollection[OwnerT: Entity, AssociateT: Entity](
    _BidirectionalAssociateCollection[AssociateT, OwnerT],
    SingleTypeEntityCollection[AssociateT],
):
    pass


class _BidirectionalMultipleTypesAssociateCollection[
    OwnerT: Entity,
    AssociateT: Entity,
](
    _BidirectionalAssociateCollection[AssociateT, OwnerT],
    MultipleTypesEntityCollection[AssociateT],
):
    pass


def resolve(*entities: Entity) -> None:
    """
    Resolve all entities' associates.

    You **MUST** call this on all entities once the resolvers you have set on them can indeed be resolved.
    """
    for entity in entities:
        for association in AssociationRegistry.get_all_associations(entity):
            association.resolve(entity)


type ToOneAssociate[EntityT: Entity] = EntityT | ToOneResolver[EntityT]
type ToZeroOrOneAssociate[EntityT: Entity] = (
    ToOneAssociate[EntityT] | ToZeroOrOneResolver[EntityT] | None
)
type ToManyAssociates[EntityT: Entity] = Iterable[EntityT] | ToManyResolver[EntityT]



"""
JSON schemas for static translations.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from betty.json.schema import Object
from betty.locale.localizable.markup import Paragraph

if TYPE_CHECKING:
    from betty.locale.localizable import ResolvableLocalizable


class StaticTranslationsSchema(Object):
    """
    A JSON Schema for :py:class:`betty.locale.localizable.static.StaticTranslations`.
    """

    def __init__(
        self,
        *,
        title: ResolvableLocalizable = "Static translations",
        description: ResolvableLocalizable | None = None,
    ):
        super().__init__(
            title=title,
            description=Paragraph(
                *([] if description is None else [description]),
                "Keys are IETF BCP-47 language tags.",
            ),
        )
        self._schema["additionalProperties"] = {
            "type": "string",
            "description": "A human-readable translation.",
        }



"""
Render HTML.
"""

from typing import Self, override

from betty.html.url import generate_urls
from betty.locale.localizable.gettext import _
from betty.locale.localizable.markup import AllEnumeration
from betty.media_type import MediaType
from betty.media_type.media_types import HTML
from betty.project import Project
from betty.render import Renderer, RendererDefinition
from betty.service.factory import Manufacturable
from betty.service.requirement.project import require_project
from betty.url import UrlGenerator

_ATTRIBUTES = ("href", "src")


@RendererDefinition(
    "html",
    label="HTML",
    description=_(
        "The values of the following HTML attributes will automatically be replaced with the URLs generated from them where possible: {attributes}"
    ).format(attributes=AllEnumeration(*_ATTRIBUTES)),
)
class Html(Manufacturable, Renderer):
    """
    .. plugin:: renderer:html.
    """

    def __init__(self, *, url_generator: UrlGenerator):
        self._url_generator = url_generator

    @override
    @classmethod
    @require_project
    async def new(cls, project: Project, /) -> Self:
        return cls(url_generator=await project.url_generator)

    @override
    @property
    def media_type(self) -> MediaType:
        return HTML

    @override
    async def render(self, content: str, /) -> str:
        return generate_urls(content, _ATTRIBUTES, url_generator=self._url_generator)



"""
Tools to create classes.
"""

from typing import Self, final, override


class Singleton:
    """
    A base class for singletons.
    """

    _instance: Self | None = None

    @final
    @override
    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance



"""
Keyed collection definitions.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, final

from betty.assertion import assert_mapping, assert_sequence
from betty.collection.keyed import KeyedCollection, MutableKeyedCollection
from betty.data.aggregate.collection import CollectionDefinition
from betty.data.indicator.selector import Element
from betty.portable import (
    CallbackPorter,
    PortableData,
    PortableMapping,
    PortableSequence,
)

if TYPE_CHECKING:
    from collections.abc import Callable

    from ty_extensions import Intersection

    from betty.data import Data
    from betty.data.aggregate.record import RecordDefinition
    from betty.locale.localizable import ResolvableLocalizable


@final
class KeyedCollectionDefinition[
    MutableKeyedCollectionT: MutableKeyedCollection = MutableKeyedCollection,
    ElementT: Element[str] = Element[str],
](CollectionDefinition[MutableKeyedCollectionT, ElementT]):
    """
    A definition for :py:class:`betty.collection.keyed.MutableKeyedCollection`.
    """

    _item: RecordDefinition[Any, ElementT]

    def __init__[KeyT, ValueT](
        self,
        /,
        cls: type[MutableKeyedCollection] | None = None,
        *,
        value: RecordDefinition[ValueT, ElementT]
        | type[Intersection[ValueT, Data[RecordDefinition[Any, ElementT]]]],
        key: ElementT,
        order_dump: bool = False,
        label: ResolvableLocalizable,
        description: ResolvableLocalizable | None = None,
        factory: Callable[[], MutableKeyedCollectionT] | None = None,
    ):
        super().__init__(
            cls=cls,
            label=label,
            description=description,
            porter=CallbackPorter(self._load, self._dump),
            item=value,
            factory=factory,
        )
        self._key = key
        self._order_dump = order_dump

    def _load(self, portable: PortableData, /) -> MutableKeyedCollection:
        if self._order_dump:
            values = assert_sequence(self._item.porter.load)(portable)
        else:
            values = [
                self._item.porter.load_key(portable_item, self._key, portable_key)
                for portable_key, portable_item in assert_mapping()(portable).items()
            ]

        loaded = self.new()
        loaded.add(*values)
        return loaded

    def _dump(self, data: KeyedCollection) -> PortableMapping | PortableSequence:
        if self._order_dump:
            return [self._item.porter.dump(value) for value in data]
        return dict(
            self._item.porter.dump_key(item_data, self._key) for item_data in data
        )



"""
Rate limit implementations.
"""

from typing import override

from aiohttp.client_reqrep import ClientRequest

from betty.http_client.rate_limit import RateLimit, RateLimitDefinition
from betty.plugin import Plugin


@RateLimitDefinition("wikipedia-action-api")
class WikipediaActionApi(RateLimit, Plugin):
    """
    .. plugin:: http-rate-limit:wikipedia-action-api.

    See https://www.mediawiki.org/wiki/API:Action_API.
    """

    @override
    def match(self, request: ClientRequest) -> bool:
        return (
            request.url.scheme in ("http", "https")
            and request.url.host is not None
            and request.url.host.endswith(".wikipedia.org")
            and request.url.path == "/w/api.php"
        )

    @override
    @property
    def limit(self) -> tuple[int, int]:
        # https://www.mediawiki.org/wiki/API:Etiquette states there are no hard limits on the Wikimedia
        # Foundation-managed Action APIs. We've taken the limit of "200 requests per second" from
        # https://www.mediawiki.org/wiki/Wikimedia_REST_API#Terms_and_conditions instead.
        return 200, 1


@RateLimitDefinition("wikipedia-rest-api")
class WikipediaRestApi(RateLimit, Plugin):
    """
    .. plugin:: http-rate-limit:wikipedia-rest-api.

    See https://www.mediawiki.org/wiki/Wikimedia_REST_API#Terms_and_conditions.
    """

    @override
    def match(self, request: ClientRequest) -> bool:
        return (
            request.url.scheme in ("http", "https")
            and request.url.host is not None
            and request.url.host.endswith(".wikipedia.org")
            and request.url.path == "/api/rest_v1"
            or request.url.path.startswith("/api/rest_v1/")
        )

    @override
    @property
    def limit(self) -> tuple[int, int]:
        return 200, 1



"""
Provide search functionality.
"""

from __future__ import annotations

import json
from asyncio import gather
from dataclasses import dataclass
from typing import TYPE_CHECKING, Generic, TypeVar, final, override

import aiofiles
from aiofiles.os import makedirs

from betty.ancestry.file import File
from betty.ancestry.has_notes import HasNotes
from betty.ancestry.person import Person
from betty.ancestry.place import Place
from betty.model import Entity, EntityDefinition
from betty.privacy import is_private

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping, Sequence

    from babel import Locale

    from betty.job import Context
    from betty.locale.localizable import Localizable
    from betty.locale.localize import Localizer
    from betty.machine_name import MachineName
    from betty.project import Project


async def generate_search_index(
    project: Project,
    result_container_template: Localizable,
    results_container_template: Localizable,
    *,
    context: Context,
) -> None:
    await gather(
        *(
            _generate_search_index_for_locale(
                project,
                result_container_template,
                results_container_template,
                locale,
                context=context,
            )
            for locale in project.configuration.locales.keys()  # noqa: SIM118
        )
    )


async def _generate_search_index_for_locale(
    project: Project,
    result_container_template: Localizable,
    results_container_template: Localizable,
    locale: Locale,
    *,
    context: Context,
) -> None:
    localizers = await project.localizers
    localizer = localizers.get(locale)
    search_index = {
        "resultContainerTemplate": result_container_template.localize(localizer),
        "resultsContainerTemplate": results_container_template.localize(localizer),
        "index": [
            {
                "entityTypeId": entry.entity_type_id,
                "text": " ".join(entry.text),
                "result": entry.result,
            }
            for entry in await Index(project, context, localizer).build()
        ],
    }
    search_index_json = json.dumps(search_index)
    www_directory_path = project.localize_www_directory(locale)
    await makedirs(www_directory_path, exist_ok=True)
    async with aiofiles.open(www_directory_path / "search-index.json", mode="w") as f:
        await f.write(search_index_json)


_EntityTypeIndexerEntityT = TypeVar(
    "_EntityTypeIndexerEntityT", bound=Entity, default=Entity, covariant=True
)


class EntityTypeIndexer(Generic[_EntityTypeIndexerEntityT]):
    def __init__(self, project: Project):
        self._project = project

    async def text(
        self, localizer: Localizer, entity: _EntityTypeIndexerEntityT
    ) -> set[str]:
        text = {entity.id.lower()}

        # Each note is owned by a single other entity, so index it as part of that entity.
        if isinstance(entity, HasNotes):
            for note in entity.notes:
                text.update(note.text.localize(localizer).lower().split())

        return text


class _FallbackIndexer(EntityTypeIndexer[Entity]):
    @override
    async def text(self, localizer: Localizer, entity: Entity) -> set[str]:
        text = await super().text(localizer, entity)
        text.update(entity.label.localize(localizer))
        return text


class _PersonIndexer(EntityTypeIndexer[Person]):
    @override
    async def text(self, localizer: Localizer, entity: Person) -> set[str]:
        text = await super().text(localizer, entity)
        for name in entity.names:
            if name.individual is not None:
                text.update(set(name.individual.lower().split()))
            if name.affiliation is not None:
                text.update(set(name.affiliation.lower().split()))
        return text


class _PlaceIndexer(EntityTypeIndexer[Place]):
    @override
    async def text(self, localizer: Localizer, entity: Place) -> set[str]:
        text = await super().text(localizer, entity)
        for name in entity.names:
            text.update(name.name.localize(localizer).lower().split())
        return text


class _FileIndexer(EntityTypeIndexer[File]):
    @override
    async def text(self, localizer: Localizer, entity: File) -> set[str]:
        text = await super().text(localizer, entity)
        text.update(entity.path.name.strip().lower().split())
        if entity.description:
            text.update(entity.description.localize(localizer).strip().lower().split())
        return text


@final
@dataclass(frozen=True)
class _Entry:
    entity_type_id: MachineName
    result: str
    text: set[str]


class Index:
    """
    Build search indexes.
    """

    def __init__(
        self,
        project: Project,
        context: Context | None,
        localizer: Localizer,
    ):
        self._project = project
        self._context = context
        self._localizer = localizer

    async def build(self) -> Sequence[_Entry]:
        """
        Build the search index.
        """
        entity_types = self._project.plugins[EntityDefinition]
        specialized_indexers: Mapping[type[Entity], EntityTypeIndexer[Entity]] = {
            File: _FileIndexer(self._project),
            Person: _PersonIndexer(self._project),
            Place: _PlaceIndexer(self._project),
        }
        return [
            entry
            for entries in await gather(
                *[
                    self._build_entities(indexer, entity_type)
                    for entity_type, indexer in specialized_indexers.items()
                ],
                *[
                    self._build_entities(
                        _FallbackIndexer(self._project), entity_type.cls
                    )
                    async for entity_type in entity_types
                    if entity_type.public_facing
                    and entity_type.cls not in specialized_indexers
                ],
            )
            for entry in entries
            if entry is not None
        ]

    async def _build_entities[EntityT: Entity](
        self, indexer: EntityTypeIndexer[EntityT], entity_type: type[EntityT]
    ) -> Iterable[_Entry | None]:
        return await gather(
            *(
                self._build_entity(indexer, entity)
                for entity in self._project.ancestry[entity_type]
            )
        )

    async def _build_entity[EntityT: Entity](
        self, indexer: EntityTypeIndexer[EntityT], entity: EntityT
    ) -> _Entry | None:
        if is_private(entity):
            return None
        text = await indexer.text(self._localizer, entity)
        if not text:
            return None
        return _Entry(entity.plugin().id, await self._render_entity(entity), text)

    async def _render_entity(self, entity: Entity) -> str:
        jinja = await self._project.jinja
        return await jinja.select_template(
            [
                f"search/result--{entity.plugin().id}.html.j2",
                "search/result.html.j2",
            ]
        ).render_async(
            document=await self._project.new_document(
                context=self._context,
                localizer=self._localizer,
            ),
            entity=entity,
        )



"""
Provide the HTML API, for generating HTML pages.
"""

from __future__ import annotations

import re
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, final
from uuid import uuid4

from markupsafe import escape

from betty.locale.localizable import resolve_localizable

if TYPE_CHECKING:
    from collections.abc import Sequence

    from betty.locale.localizable import Localizable, ResolvableLocalizable


class CssProvider(ABC):
    """
    Provide CSS for HTML pages.
    """

    @abstractmethod
    async def get_public_css_paths(self) -> Sequence[str]:
        """
        The URL-generatable resources of the CSS files to include in each HTML page.
        """


class JsProvider(ABC):
    """
    Provide JavaScript for HTML pages.
    """

    @abstractmethod
    async def get_public_js_paths(self) -> Sequence[str]:
        """
        The URL-generatable resources of the JS files to include in each HTML page.
        """


@final
class NavigationLink:
    """
    A navigation link.
    """

    def __init__(self, url: ResolvableLocalizable, label: ResolvableLocalizable):
        self._url = resolve_localizable(url)
        self._label = resolve_localizable(label)

    @property
    def url(self) -> Localizable:
        """
        The URL the link points to.
        """
        return self._url

    @property
    def label(self) -> Localizable:
        """
        The human-readable short link label.
        """
        return self._label


class NavigationLinkProvider:
    """
    Provide navigation links for HTML pages.
    """

    def primary_navigation_links(self) -> Sequence[NavigationLink]:
        """
        The primary navigation links.
        """
        return ()

    def secondary_navigation_links(self) -> Sequence[NavigationLink]:
        """
        The secondary navigation links.
        """
        return ()


_paragraph_re = re.compile(r"(?:\r\n|\r|\n){2,}")


def newlines_to_paragraphs(text: str) -> str:
    """
    Convert newlines to <p> and <br> tags.
    """
    return "\n\n".join(
        "<p>{}</p>".format(paragraph.replace("\n", "<br>\n"))
        for paragraph in _paragraph_re.split(text)
    )


def plain_text_to_html(text: str) -> str:
    """
    Convert plain text to HTML.
    """
    return newlines_to_paragraphs(escape(text))


def generate_html_id() -> str:
    """
    Generate a unique HTML ID.
    """
    return f"betty-generated--{uuid4()}"



from __future__ import annotations  # noqa: D100

from typing import TYPE_CHECKING, Any
from urllib.parse import urlparse

from betty.ancestry.event import Event
from betty.ancestry.person import Person
from betty.ancestry.place import Place
from betty.ancestry.source import Source
from betty.assertion import assert_locale, assert_path, assert_str
from betty.extension import Extension, ExtensionDefinition, ExtensionManufacturer
from betty.extension.deriver import Deriver
from betty.extension.gramps import Gramps
from betty.extension.gramps.data import FamilyTree, GrampsConfiguration
from betty.extension.http_api_doc import HttpApiDoc
from betty.extension.maps import Maps
from betty.extension.privatizer import Privatizer
from betty.extension.raspberry_mint import RaspberryMint
from betty.extension.raspberry_mint.data import RaspberryMintConfiguration
from betty.extension.raspberry_mint.default import regional_content
from betty.extension.trees import Trees
from betty.extension.webpack import Webpack
from betty.extension.wiki import Wiki
from betty.locale import DEFAULT_LOCALE_TAG, to_language_tag
from betty.locale.localizable.gettext import _
from betty.locale.localizable.static import StaticTranslations
from betty.machine_name import MachineName
from betty.portable.file import dump_file
from betty.project.data import ProjectConfiguration
from betty.typing import Void

if TYPE_CHECKING:
    from collections.abc import MutableSequence, Sequence
    from pathlib import Path

    from babel import Locale

    from betty.app import App
    from betty.locale.localizable import Localizable
    from betty.plugin.factory import ResolvablePluginManufacturer
    from betty.user import User


async def new(app: App) -> None:
    """
    Create a new project.
    """
    localizers = await app.localizers

    configuration_file_path = await app.user.ask_input(
        _("Where do you want to save your project's configuration file?"),
        assertion=_assert_project_configuration_file_path,
    )

    locales = [
        await app.user.ask_input(
            _(
                "Which language should your project site be generated in? Enter a language code."
            ),
            default=DEFAULT_LOCALE_TAG,
            assertion=assert_locale(),
        )
    ]
    while await app.user.ask_confirmation(_("Do you want to add another locale?")):
        locales.append(
            await app.user.ask_input(
                _(
                    "Which language should your project site be generated in? Enter a language code."
                ),
                assertion=assert_locale(),
            )
        )

    extensions: MutableSequence[
        ResolvablePluginManufacturer[ExtensionDefinition, Extension]
    ] = [
        Deriver,
        HttpApiDoc,
        Maps,
        Privatizer,
        ExtensionManufacturer(
            RaspberryMint,
            RaspberryMintConfiguration(
                regional_content=regional_content(
                    localizers=[localizers.get(locale) for locale in locales]
                )
            ),
        ),
        Trees,
        # Enable the Webpack extension explicitly for the test's mock to work.
        Webpack,
        Wiki,
    ]

    title = await _user_input_static_translations(
        app.user, locales, _("What is your project called in {locale}?")
    )

    name = await app.user.ask_input(
        _("What is your project's machine name?"),
        default=MachineName.machinify(title.localize(localizers.get(locales[0])))
        or Void,
        assertion=MachineName,
    )

    author = await _user_input_static_translations(
        app.user, locales, _("What is the project author called in {locale}?")
    )

    url = await app.user.ask_input(
        _("At which URL will your site be published?"),
        default="https://example.com",
        assertion=_assert_url,
    )

    if await app.user.ask_confirmation(_("Do you want to load a Gramps family tree?")):
        extensions.append(
            ExtensionManufacturer(
                Gramps,
                GrampsConfiguration(
                    family_trees=[
                        FamilyTree(
                            await app.user.ask_input(
                                _(
                                    "What is the path to your exported Gramps family tree file?"
                                ),
                                assertion=assert_path(),
                            )
                        )
                    ]
                ),
            )
        )

    configuration = ProjectConfiguration(
        author=author,
        locales=locales,
        entity_types=[
            Person,
            Event,
            Place,
            Source,
        ],
        extensions=extensions,
        name=name,
        title=title,
        url=url,
    )
    await dump_file(
        configuration.data().porter.dump(configuration), configuration_file_path
    )
    await app.user.message_information(
        _("Saved your project to {configuration_file}.").format(
            configuration_file=str(configuration_file_path)
        )
    )


def _assert_project_configuration_file_path(value: Any) -> Path:
    configuration_file_path = assert_path()(value)
    if not configuration_file_path.suffix:
        configuration_file_path /= "betty.yaml"
    return configuration_file_path


def _assert_url(value: Any) -> str:
    url = assert_str()(value)
    parsed_url = urlparse(url)
    scheme = parsed_url.scheme or "http"
    return f"{scheme}://{parsed_url.netloc}{parsed_url.path}"


async def _user_input_static_translations(
    user: User, locales: Sequence[Locale], question: Localizable
) -> StaticTranslations:
    return StaticTranslations(
        {
            locale: await user.ask_input(
                question.format(
                    locale=locale.get_display_name() or to_language_tag(locale)
                )
            )
            for locale in locales
        }
    )



"""Provide information about (this version of) Betty."""

from __future__ import annotations

from importlib import metadata

_DEV_VERSION = "0.0.0"


VERSION = metadata.version("betty")
"""
The current Betty installation's version.
"""

_VERSION_MAJOR_PARTS = VERSION.split(".")
VERSION_MAJOR = (
    f"0.{_VERSION_MAJOR_PARTS[1]}"
    if _VERSION_MAJOR_PARTS[0] == "0"
    else _VERSION_MAJOR_PARTS[0]
)
"""
The current Betty installation's major version.

Use this instead of :py:const:`betty.about.VERSION` in public situations, where we do not want to include minor version
information because that may disclose security issues.
"""


IS_STABLE = VERSION != _DEV_VERSION
"""
Whether the current Betty installation is a stable version.
"""


IS_DEVELOPMENT = not IS_STABLE
"""
Whether the current Betty installation is an unstable development version.
"""


VERSION_LABEL = "development" if IS_DEVELOPMENT else VERSION
"""
The human-readable label for the current Betty installation's version.
"""


VERSION_MAJOR_LABEL = "development" if IS_DEVELOPMENT else VERSION_MAJOR
"""
The human-readable label for the current Betty installation's major version.

Use this instead of :py:const:`betty.about.VERSION_LABEL` in public situations, where we do not want to include minor
version information because that may disclose security issues.
"""



"""
Provide asynchronous programming utilities.
"""

from __future__ import annotations

from inspect import isawaitable
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Awaitable


async def resolve_await[T](value: Awaitable[T] | T) -> T:
    """
    Return a value, but await it first if it is awaitable.
    """
    if isawaitable(value):
        return await value  # ty:ignore[invalid-return-type]
    return value



"""
Provide Betty's ancestry place types.
"""

from __future__ import annotations

from typing import final

from betty.classtools import Singleton
from betty.locale.localizable.gettext import _, ngettext
from betty.place_type import PlaceType, PlaceTypeDefinition


@final
@PlaceTypeDefinition(
    "borough",
    label=_("Borough"),
    label_plural=_("Boroughs"),
    label_countable=ngettext("{count} borough", "{count} boroughs"),
)
class Borough(PlaceType):
    """
    .. plugin:: place-type:borough.
    """


@final
@PlaceTypeDefinition(
    "building",
    label=_("Building"),
    label_plural=_("Buildings"),
    label_countable=ngettext("{count} building", "{count} buildings"),
)
class Building(PlaceType):
    """
    .. plugin:: place-type:building.
    """


@final
@PlaceTypeDefinition(
    "cemetery",
    label=_("Cemetery"),
    label_plural=_("Cemeteries"),
    label_countable=ngettext("{count} cemetery", "{count} cemeteries"),
)
class Cemetery(PlaceType):
    """
    .. plugin:: place-type:cemetery.
    """


@final
@PlaceTypeDefinition(
    "city",
    label=_("City"),
    label_plural=_("Cities"),
    label_countable=ngettext("{count} city", "{count} cities"),
)
class City(PlaceType):
    """
    .. plugin:: place-type:city.
    """


@final
@PlaceTypeDefinition(
    "country",
    label=_("Country"),
    label_plural=_(""),
    label_countable=ngettext("{count} ", "{count} "),
)
class Country(PlaceType):
    """
    .. plugin:: place-type:country.
    """


@final
@PlaceTypeDefinition(
    "county",
    label=_("County"),
    label_plural=_("Counties"),
    label_countable=ngettext("{count} county", "{count} counties"),
)
class County(PlaceType):
    """
    .. plugin:: place-type:county.
    """


@final
@PlaceTypeDefinition(
    "department",
    label=_("Department"),
    label_plural=_("Departments"),
    label_countable=ngettext("{count} department", "{count} departments"),
)
class Department(PlaceType):
    """
    .. plugin:: place-type:department.
    """


@final
@PlaceTypeDefinition(
    "district",
    label=_("District"),
    label_plural=_("Districts"),
    label_countable=ngettext("{count} district", "{count} districts"),
)
class District(PlaceType):
    """
    .. plugin:: place-type:district.
    """


@final
@PlaceTypeDefinition(
    "farm",
    label=_("Farm"),
    label_plural=_("Farms"),
    label_countable=ngettext("{count} farm", "{count} farms"),
)
class Farm(PlaceType):
    """
    .. plugin:: place-type:farm.
    """


@final
@PlaceTypeDefinition(
    "hamlet",
    label=_("Hamlet"),
    label_plural=_("Hamlets"),
    label_countable=ngettext("{count} hamlet", "{count} hamlets"),
)
class Hamlet(PlaceType):
    """
    .. plugin:: place-type:hamlet.
    """


@final
@PlaceTypeDefinition(
    "locality",
    label=_("Locality"),
    label_plural=_("Localities"),
    label_countable=ngettext("{count} locality", "{count} localities"),
)
class Locality(PlaceType):
    """
    .. plugin:: place-type:locality.
    """


@final
@PlaceTypeDefinition(
    "municipality",
    label=_("Municipality"),
    label_plural=_("Municipalities"),
    label_countable=ngettext("{count} municipality", "{count} municipalities"),
)
class Municipality(PlaceType):
    """
    .. plugin:: place-type:municipality.
    """


@final
@PlaceTypeDefinition(
    "neighborhood",
    label=_("Neighborhood"),
    label_plural=_("Neighborhoods"),
    label_countable=ngettext("{count} neighborhood", "{count} neighborhoods"),
)
class Neighborhood(PlaceType):
    """
    .. plugin:: place-type:neighborhood.
    """


@final
@PlaceTypeDefinition(
    "number",
    label=_("Number"),
    label_plural=_("Numbers"),
    label_countable=ngettext("{count} number", "{count} numbers"),
    description=_("A place number, such as a house or flat number."),
)
class Number(PlaceType):
    """
    .. plugin:: place-type:number.
    """


@final
@PlaceTypeDefinition(
    "parish",
    label=_("Parish"),
    label_plural=_("Parishes"),
    label_countable=ngettext("{count} parish", "{count} parishes"),
)
class Parish(PlaceType):
    """
    .. plugin:: place-type:parish.
    """


@final
@PlaceTypeDefinition(
    "province",
    label=_("Province"),
    label_plural=_("Provinces"),
    label_countable=ngettext("{count} province", "{count} provinces"),
)
class Province(PlaceType):
    """
    .. plugin:: place-type:province.
    """


@final
@PlaceTypeDefinition(
    "region",
    label=_("Region"),
    label_plural=_("Regions"),
    label_countable=ngettext("{count} region", "{count} regions"),
)
class Region(PlaceType):
    """
    .. plugin:: place-type:region.
    """


@final
@PlaceTypeDefinition(
    "state",
    label=_("State"),
    label_plural=_("States"),
    label_countable=ngettext("{count} state", "{count} states"),
)
class State(PlaceType):
    """
    .. plugin:: place-type:state.
    """


@final
@PlaceTypeDefinition(
    "street",
    label=_("Street"),
    label_plural=_("Streets"),
    label_countable=ngettext("{count} street", "{count} streets"),
)
class Street(PlaceType):
    """
    .. plugin:: place-type:street.
    """


@final
@PlaceTypeDefinition(
    "town",
    label=_("Town"),
    label_plural=_("Towns"),
    label_countable=ngettext("{count} town", "{count} towns"),
)
class Town(PlaceType):
    """
    .. plugin:: place-type:town.
    """


@final
@PlaceTypeDefinition(
    "unknown",
    label=_("Unknown"),
    label_plural=_("Unknowns"),
    label_countable=ngettext("{count} unknown", "{count} unknowns"),
)
class Unknown(PlaceType, Singleton):
    """
    .. plugin:: place-type:unknown.
    """


@final
@PlaceTypeDefinition(
    "village",
    label=_("Village"),
    label_plural=_("Villages"),
    label_countable=ngettext("{count} village", "{count} villages"),
)
class Village(PlaceType):
    """
    .. plugin:: place-type:village.
    """



"""
Map content.
"""

from typing import Self, override

from betty.ancestry.event import Event
from betty.ancestry.person import Person
from betty.ancestry.place import Place
from betty.content import ContentDefinition
from betty.content.contents import Template, TemplateBuild
from betty.document import Document
from betty.extension.maps import Maps
from betty.locale.localizable.gettext import _
from betty.project import Project
from betty.service.factory import Manufacturable
from betty.service.requirement.extension import require_extension
from betty.service.requirement.project import require_project


@ContentDefinition("maps-map", label=_("Map"))
class Map(Template, Manufacturable):
    """
    An interactive map.

    .. plugin:: content:maps-map
    """

    @override
    @classmethod
    @require_project
    async def new(cls, project: Project, /) -> Self:
        await require_extension(Maps, project)
        return cls(jinja=await project.jinja)

    @override
    async def build_template(self, document: Document) -> TemplateBuild:
        places = []
        if isinstance(document.resource, Event) and document.resource.place:
            places.append(document.resource.place)
        elif isinstance(document.resource, Person):
            places.extend(
                presence.event.place
                for presence in document.resource.presences
                if presence.public
                and presence.event.public
                and presence.event.place
                and presence.event.place.public
            )
        elif isinstance(document.resource, Place):
            places.append(
                document.resource,
            )
        if places:
            return "component/maps/map.html.j2", {
                "places": places,
            }
        return None


@ContentDefinition("maps-attribution", label=_("Map attribution"))
class Attribution(Template, Manufacturable):
    """
    The attribution for an interactive map.

    .. plugin:: content:maps-attribution
    """

    @override
    @classmethod
    @require_project
    async def new(cls, project: Project, /) -> Self:
        await require_extension(Maps, project)
        return cls(jinja=await project.jinja)

    @override
    async def build_template(self, document: Document) -> TemplateBuild:
        return "component/maps/attribution.html.j2"



"""
Configuration for ordered plugins.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from betty.data.aggregate.collection.sequence import SequenceDefinition
from betty.locale.localizable.gettext import _
from betty.machine_name import MachineName
from betty.plugin import PluginDefinition, ResolvableId, resolve_id
from betty.plugin.data import PluginDefinitionConfiguration
from betty.property.collection.sequence import SequenceProperty

if TYPE_CHECKING:
    from collections.abc import Iterable


class OrderedPluginDefinitionConfiguration[PluginDefinitionT: PluginDefinition](
    PluginDefinitionConfiguration[PluginDefinitionT]
):
    """
    Configure a :py:class:`betty.plugin.ordered.OrderedPluginDefinition`.

    .. data:: betty.plugin.config.ordered:OrderedPluginDefinitionConfiguration
    """

    comes_before = SequenceProperty(
        SequenceDefinition(cls=list, label=_("Comes before"), value=MachineName),
        omit_load=True,
        omit_dump=lambda data: not len(data),
    )
    comes_after = SequenceProperty(
        SequenceDefinition(cls=list, label=_("Comes after"), value=MachineName),
        omit_load=True,
        omit_dump=lambda data: not len(data),
    )

    def __init__(
        self,
        comes_before: Iterable[ResolvableId] | None = None,
        comes_after: Iterable[ResolvableId] | None = None,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        if comes_before is not None:
            self.comes_before = map(resolve_id, comes_before)
        if comes_after is not None:
            self.comes_after = map(resolve_id, comes_after)



"""
Jobs.
"""

from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING, override

from betty.ancestry.person import Person
from betty.job import Job
from betty.locale.localizable.gettext import _, ngettext
from betty.privacy import HasPrivacy

if TYPE_CHECKING:
    from collections.abc import MutableMapping, MutableSequence

    from ty_extensions import Intersection

    from betty.job.scheduler import Scheduler
    from betty.machine_name import MachineName
    from betty.model import Entity
    from betty.project import Project


class PrivatizeAncestry(Job):
    """
    Privatize an ancestry.
    """

    def __init__(self, dependencies: set[str] | None = None, *, project: Project):
        super().__init__(self.id_for(), dependencies=dependencies)
        self._project = project

    @classmethod
    def id_for(cls) -> str:
        """
        Get the job ID.
        """
        return "privatizer:privatize"

    @override
    async def do(self, scheduler: Scheduler, /) -> None:
        await self._project.app.localizer
        user = self._project.app.user

        newly_privatized: MutableMapping[MachineName, int] = defaultdict(lambda: 0)
        entities: MutableSequence[Intersection[HasPrivacy, Entity]] = []
        for entity in self._project.ancestry:
            if isinstance(entity, HasPrivacy):
                entities.append(entity)
                if entity.private:
                    newly_privatized[entity.plugin().id] -= 1

        for entity in entities:
            await self._project.privatizer.privatize(entity)

        for entity in entities:
            if entity.private:
                newly_privatized[entity.plugin().id] += 1

        if newly_privatized[Person.plugin().id] > 0:
            await user.message_information_details(
                _(
                    "Privatized {count} people because they are likely still alive."
                ).format(
                    count=str(newly_privatized[Person.plugin().id]),
                )
            )
        for entity_type_id in set(newly_privatized) - {Person.plugin().id}:
            if newly_privatized[entity_type_id] > 0:
                await user.message_information_details(
                    ngettext(
                        'Privatized {count} "{entity_type_id}" entity, because it is associated with private information.',
                        'Privatized {count} "{entity_type_id}" entities, because they are associated with private information.',
                        newly_privatized[entity_type_id],
                    ).format(
                        entity_type_id=entity_type_id,
                    )
                )
